// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file PetscSolverGlobalMatrix.h */

#ifndef AMDIS_PETSC_SOLVER_GLOBAL_MATRIX_H
#define AMDIS_PETSC_SOLVER_GLOBAL_MATRIX_H

#include <boost/tuple/tuple.hpp>
#include "AMDiS_fwd.h"
#include "parallel/MatrixNnzStructure.h"
#include "parallel/PetscSolver.h"

namespace AMDiS {

  using namespace std;


  class PetscSolverGlobalMatrix : public PetscSolver
  {
  public:
    PetscSolverGlobalMatrix()
      : PetscSolver(),
	petscSolVec(PETSC_NULL),
	lastMeshNnz(0),
	zeroStartVector(false),
	alwaysCreateNnzStructure(false)
    {
      Parameters::get("parallel->use zero start vector", zeroStartVector);
      Parameters::get("parallel->always create nnz structure", 
		      alwaysCreateNnzStructure);
    }

    void fillPetscMatrix(Matrix<DOFMatrix*> *mat);

    /// This function is just a small wrapper that creates a 1x1 matrix that
    /// contains exactly one DOFMatrix and than calls \ref fillPetscMatrix
    void fillPetscMatrix(DOFMatrix *mat);

    void fillPetscMatrixWithCoarseSpace(Matrix<DOFMatrix*> *mat);

    void fillPetscRhs(SystemVector *vec);

    void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo);    

    void solveGlobal(Vec &rhs, Vec &sol);

    void destroyMatrixData();

    void destroyVectorData();

  protected:
    bool checkMeshChange(Matrix<DOFMatrix*> *mat, bool localMatrix = false);

    void createFieldSplit(PC pc);

    virtual void initPreconditioner(PC pc);

    virtual void exitPreconditioner(PC pc);

    /// Takes a DOF matrix and sends the values to the global PETSc matrix.
    void setDofMatrix(DOFMatrix* mat, int nRowMat = 0, int nColMat = 0);

    /// Takes a DOF vector and sends its values to a given PETSc vector.
    void setDofVector(Vec vecInterior,
		      Vec vecCoarse,
		      DOFVector<double>* vec, 
		      int nRowVec, bool rankOnly = false);

    void removeDirichletBcDofs(Matrix<DOFMatrix*> *mat);

    void removeDirichletBcDofs(SystemVector *vec);

    inline void setDofVector(Vec vecInterior,
			     DOFVector<double>* vec, 
			     int nRowVec, bool rankOnly = false)
    {
      setDofVector(vecInterior, PETSC_NULL, vec, nRowVec, rankOnly);
    }


  protected:
    MatrixNnzStructure nnzInterior, nnzCoarse;

    MatrixNnzStructure nnzIntCoarse, nnzCoarseInt;

    Vec petscSolVec;

    /// Stores the mesh change index of the mesh the nnz structure was created for.
    /// Therefore, if the mesh change index is higher than this value, we have to create
    /// a new nnz structure for PETSc matrices, because the mesh has been changed and
    /// therefore also the assembled matrix structure.
    int lastMeshNnz;

    bool zeroStartVector;

    /// If this variable is set to true, the non-zero matrix structure is
    /// created each time from scratch by calling \ref createPetscNnzStrcuture.
    /// This can be necessary if the number of non-zeros in the matrix varies
    /// though the mesh does not change. This may happen if there are many
    /// operators using DOFVectors from old timestep containing many zeros due to
    /// some phase fields.
    bool alwaysCreateNnzStructure;
  };


}

#endif

